#!/usr/bin/env python
# File name: make_model.py
# Author: Matt Robinson
# Date created: 5/24/2017
# Date last modified: 9/05/2017
# Python Version: 3.6
"""
Description:
This script uses the alignment file, alignment.ali, to make a homology model that includes the previously missing residues. 
In this case, the original PDB structure is the template while the full sequence is the target of the model. 

Usage: python make_model.py pdbfile.pdb [options]

Input Arguments:
[optional]
-a, --automodel
    The simplest method for simple comparitive modeling. 
    Will not give great results but suggested when many chain breaks are present. [default: True]

-f, --fixed_automodel
    Builds an automodel and keeps the non-missing residues fixed,
    whereas they can move in the other methods. [default: False]
    
-l, --loopmodel 
    Builds a model by refining the loop with the missing residues. 
    Suggested when have one small chain break in the PDB. [default: False]

Output: A set of PDB files (number depends on the chosen method)

Note: The alignment file must be in the same directory as this script.
"""
import sys
import argparse
import os
from modeller import *
from modeller.automodel import *    # Load the automodel class

def main():

    parser = argparse.ArgumentParser(
        prog='make_model.py',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=
    """
    This script uses the alignment file, alignment.ali, 
    to make a homology model that includes the previously missing residues. 
    In this case, the original PDB structure is the template
    while the full sequence is the target of the model. 

    @author: Matt Robinson, matthew.robinson@yale.edu
             William L. Jorgensen lab, Yale University

    For simple automodel,
    Usage: python make_model.py -p pdbfile.pdb -a

    For fixed automodel,
    Usage: python make_model.py -p pdbfile.pdb -f

    For loopmodel,
    Usage: python make_model.py -p pdbfile.pdb -l
    
    REQUIREMENTS:
    Preferably Anaconda python 3 with following modules:
        argparse
        modeller    
    """
    )

    parser.add_argument(
        "-p", "--pdb", help="full path of the pdb file with .pdb file descriptor")
    parser.add_argument(
        "-a", "--automodel", help="the simplest method for simple comparitive modeling", action="store_true")
    parser.add_argument(
        "-f", "--fixed_automodel", help="builds an automodel and keeps the non-missing residues fixed", action="store_true")
    parser.add_argument(
        "-l", "--loopmodel", help="builds a model by refining the loop with the missing residues", action="store_true")

    args = parser.parse_args()

    # call the model function
    model(args.pdb, args.automodel, args.fixed_automodel, args.loopmodel)

def model(pdb_file, a, f, l):

    log.verbose()
    env = environ()

    # directories for input atom files
    env.io.atom_files_directory = ['.', '../atom_files']
    # Read in HETATM records from template PDBs
    env.io.hetatm = True


    #first need to get PDB data so can find missing residues
    pdb_id = os.path.splitext(os.path.basename(pdb_file))[0]

    with open(pdb_file) as pdb:
        pdb_file_data = pdb.readlines()

    with open('./' + pdb_id + '_alignment.ali') as aln_file:
        aln_data = aln_file.readlines()

    # do modeller stuff
    pdb_seq = get_pdb_seq(aln_data)

    if (f):
        #make the string for selecting the residues to move (the missing residues)
        selection_str = make_sel_str(find_missing_residues(pdb_seq))
        #selection_str = "self.residue_range('647:B', '648:B')"
        print(selection_str)

        #build the model
        class MyModel(automodel):
            def select_atoms(self):
                #select only the missing residues
                return eval('selection(' + selection_str + ')') #need to use eval b/c sel_str is str

        a = MyModel(env, alnfile = (pdb_id + '_alignment.ali'),
                    knowns = pdb_file, sequence = pdb_id + '_fill',
                    assess_methods=(assess.DOPE, assess.GA341))

        #build only one model
        a.starting_model= 1
        a.ending_model  = 1

        a.make()

    elif (l):
        a = loopmodel(env, alnfile = (pdb_id + '_alignment.ali'),
                      knowns = pdb_file, sequence = pdb_id + '_fill',
                      assess_methods=(assess.DOPE, assess.GA341))
        a.starting_model = 1
        a.ending_model   = 1

        a.loop.starting_model = 1
        a.loop.ending_model   = 2
        a.loop.md_level       = refine.fast

        a.make()

    else:
        a = automodel(env, alnfile = (pdb_id + '_alignment.ali'),
                      knowns = pdb_file, sequence = pdb_id + '_fill',
                      assess_methods=(assess.DOPE, assess.GA341))
        a.starting_model= 1
        a.ending_model  = 1

        a.make()

def find_missing_residues(pdb_seq):
    ##first delete all / from string
    #pdb_seq = pdb_seq.replace("/","")
    missing_res_lists = []
    res_number = 0 

    #create a list for holding chain labels
    chain_labels = ['A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z']


    #first split the chains
    chains_l = pdb_seq.split('/')
    # go through and find every dash in string and note index
    chain_number = 0
    res_idx = 0
    for pdb_seq in chains_l:

        missing_res_l = []
        for i in range(len(pdb_seq)):
            if (pdb_seq[i] == '-'):
                missing_res_l.append(res_idx+1) #+1 b/c Modeller starts at 1
            res_idx = res_idx + 1

        #Split missing_res into seperate loops
        chain_missing_res_lists = [] 

        start_idx = 0
        for i in range(len(missing_res_l)-1): 
            #split if not consecutive residues
            if (not(missing_res_l[i] + 1 == missing_res_l[i+1])):
                chain_missing_res_lists.append(missing_res_l[start_idx:i+1])   
                start_idx = i+1   
            #also split if end of list
            if (i==len(missing_res_l)-2):
                chain_missing_res_lists.append(missing_res_l[start_idx:i+2])

        print(chain_missing_res_lists)

        #if only one chain, don't need to add residue numbers. Just make string
        if (len(chains_l)==1):
            for lst in chain_missing_res_lists:
                lst = [str(i) for i in lst]
                missing_res_lists.append(lst)

        #go over all lists and add chain identifier if there is more than one chain 
        else:
            for lst in chain_missing_res_lists:
                lst = [str(i) for i in lst]
                lst_idx = 0
                for atom_str in lst:
                    lst[lst_idx] = atom_str + ':' + chain_labels[chain_number]
                    lst_idx = lst_idx + 1
                missing_res_lists.append(lst)

        chain_number = chain_number + 1


    print(missing_res_l)
    print(missing_res_lists)
    return missing_res_lists            

def make_sel_str(missing_res_ls):
    sel_str = ''
    for l in missing_res_ls:
        # need to add offset since Modeller numbering is different
        first_idx = l[0]
        last_idx = l[-1]
        # make the str to use as an argument
        sel_str = sel_str + "self.residue_range('" + str(first_idx) + "', '" + str(last_idx) + "'),"
    # take off the final comma
    sel_str = sel_str[0:-1]
    return sel_str

def get_pdb_seq(aln_data):
    pdb_seq = ""
    seq_data = aln_data[3:]
    for line in seq_data:
        line = line.rstrip()
        if (line[0]=='>'):
            break
        pdb_seq = pdb_seq + line
    #remove the break character '/'
    #pdb_seq = re.sub('/','',pdb_seq)
    #print(pdb_seq)
    return pdb_seq


if __name__ == "__main__":

    main()




#!/usr/bin/env python
# File name: pandas_make_alignment.py
# Author: Matt Robinson
# Date created: 06/05/2017
# Date last modified: 09/05/2017
# Python Version: 3.6
"""
Description:
Makes the .ali alignment file needed to construct the homology model.
Does this by aligning sequence from original pdb to fasta sequence.

Usage: python make_alignment.py -p pdbfile.pdb -s pdbseq.seq -f fastafile.fasta
"""
import sys
import re
import os 
from biopandas.pdb import PandasPdb
import pandas as pd
import argparse

#list so can make chain dictionaries later
chain_labels_l = ['A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z'] 

def main():

    parser = argparse.ArgumentParser(
        prog='make_alignment.py',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=
    """
    Script to align sequence from original pdb to fasta sequence.
    Returns the .ali alignment file needed to construct the homology model.

    @author: Matt Robinson, matthew.robinson@yale.edu
             William L. Jorgensen Lab, Yale University
    
    Usage: python make_alignment.py -p pdbfile.pdb -s pdbseq.seq -f fastafile.fasta

    REQUIREMENTS:
    Preferably Anaconda python 3 with following modules:
        biopandas
        pandas 
        argparse
    """
    )

    parser.add_argument(
        "-p", "--pdb", help="full path of the pdb file with .pdb file descriptor")
    parser.add_argument(
        "-s", "--seq", help="full path of sequence file with .seq file descriptor")
    parser.add_argument(
        "-f", "--fasta", help="full path of fasta file with .fasta file descriptor")

    args = parser.parse_args()

    align(args.pdb, args.seq, args.fasta)

def align(pdb_file, seq_file, fasta_file):

    #Use biopandas to create pandas dataframes of PDB info
    atom_df = make_atom_df(pdb_file)[0]
    others_df = make_atom_df(pdb_file)[1]

    #read in the sequence data
    with open(seq_file) as pdb_seq_file:
        pdb_seq_data = pdb_seq_file.readlines()

    with open(fasta_file) as fasta_file:
        full_seq_data = fasta_file.readlines()

    #get the actual sequences
    pdb_seq = get_pdb_seq(pdb_seq_data)
    full_seq = get_full_seq(full_seq_data)

    #need to trim off ligand of pdb_seq
    ligand_str = trim_ligand_seq(pdb_seq)[1]
    pdb_seq = trim_ligand_seq(pdb_seq)[0]

    #break the two seqs up chain by chain
    pdb_seq_l = break_into_chains(pdb_seq)
    full_seq_l = break_into_chains(full_seq)

    #check if they have the same number of chains
    if not(len(pdb_seq_l) == len(full_seq_l)):
            print("Fasta sequence has incorrect number of chains")
            sys.exit(2)

    #make chain dictionaries, where keys are the chain ids
    pdb_seq_chain_dict = make_chain_dict(pdb_seq_l)
    full_seq_chain_dict = make_chain_dict(full_seq_l)

    #get the translation table for chain id's 
    id_dict = make_atom_df(pdb_file)[2]

    #get a dict of all the missing residues, keys are chain id, values are lists
    missing_res_dict = make_missing_res_dict(others_df, id_dict)

    #get a dict of just the res numbers
    missing_res_num_dict = make_res_num_dict(missing_res_dict)
    #get a dict of just the res letters (1 letter code)
    missing_res_letter_dict = make_res_letter_dict(missing_res_dict)

    #delete empty chains from pdb and fasta dicts
    ### need to test these###
    pdb_seq_chain_dict = {key:pdb_seq_chain_dict[key] for key in pdb_seq_chain_dict if pdb_seq_chain_dict[key]!=[]}
    full_seq_chain_dict = {key:full_seq_chain_dict[key] for key in full_seq_chain_dict if full_seq_chain_dict[key]!=[]}

    #loop over the chains
    for key in pdb_seq_chain_dict:
        print('chain:' + key)

        #get the seqs of this chain 
        chain_pdb_seq = pdb_seq_chain_dict[key]
        chain_full_seq = full_seq_chain_dict[key]
        flank_res_num_l = get_flank_res_num_l(missing_res_num_dict[key])

        #trim the full sequence so that the non-crystallized ends are removed
        #first make list of the chains
        keys_list = list(pdb_seq_chain_dict.keys())
        #also check if it is last sequence and need to take off '* character'
        #print(chain_full_seq)
        if (not(key == keys_list[-1])):
            chain_full_seq = trim_full_seq(chain_pdb_seq, chain_full_seq)
        else:
            chain_full_seq = trim_full_seq(chain_pdb_seq[0:-1], chain_full_seq[0:-1])
            chain_full_seq = chain_full_seq + '*' #add back the '*' end character
        #print(chain_full_seq)

        #loop over the gaps
        for l in flank_res_num_l:

            print('loop:')
            print(l)

            #get the seq residues flanking the break and the residues b/w which break occurs
            flank_seq, gap_res, seq_before, seq_after = get_flank_seq(atom_df, l, key) 

            # go to next loop if was an index error on the last one
            if flank_seq == '':
                continue

            #make into one letter seq to work with fasta file
            flank_seq = make_one_letter(flank_seq)
            gap_res = make_one_letter(gap_res)
            seq_before = make_one_letter(seq_before)
            seq_after = make_one_letter(seq_after)

            print('seq:' + flank_seq)
            print('gap res:' + gap_res)
            print('seq_before:' + seq_before)
            print('seq_after:' + seq_after)

            chain_pdb_seq = add_gaps(seq_before,seq_after,chain_full_seq,chain_pdb_seq) #perhaps should be moved so dashes don't interfere

        pdb_seq_chain_dict[key] = chain_pdb_seq
        full_seq_chain_dict[key] = chain_full_seq

    # put seqs back together
    pdb_seq = '/'.join(list(pdb_seq_chain_dict.values()))
    full_seq = '/'.join(list(full_seq_chain_dict.values()))

    #add back the ligands
    pdb_seq = pdb_seq[0:-1] + ligand_str + '*'
    full_seq = full_seq[0:-1] + ligand_str + '*'

    #split both of the seqs into 75 character substrings for .pir format
    pdb_seq_substr = split_sequence(pdb_seq)
    full_seq_substr = split_sequence(full_seq)

    #take off file ending to get pdb id
    pdb_id = os.path.splitext(os.path.basename(pdb_file))[0]

    #write the output file into the local directory
    aln_file = open((pdb_id + "_alignment.ali"),"w+")

    for line in pdb_seq_data[0:3]:
        aln_file.write(line)

    for substr in pdb_seq_substr:
        aln_file.write(substr + '\n')

    aln_file.write('>P1;' + pdb_id + '_fill' + '\n') #header for full seq
    aln_file.write('sequence:::::::::'+'\n')

    for substr in full_seq_substr[0:-1]:
        aln_file.write(substr + '\n')
    aln_file.write(full_seq_substr[-1]) #don't want to end with newline character

    aln_file.close()

def make_atom_df(pdb_file):
    #Use biopandas to create pandas dataframes of PDB info
    ppdb = PandasPdb()
    ppdb.read_pdb(pdb_file) 
    atom_df = ppdb.df['ATOM']
    hetatm_df = ppdb.df['HETATM']
    others_df = ppdb.df['OTHERS'] #pandas dataframe with all the non-atom info in PDB file

    #get the last atom number before the ligands
    last_atom_num = atom_df['atom_number'].iloc[-1]

    #make atom_df so it contains both atoms and hetatms
    atom_df = pd.concat([hetatm_df, atom_df])
    #sort based on atom_number so it's in order
    atom_df = atom_df.sort_values(by=['atom_number'])
    #take out water molecules
    atom_df = atom_df.loc[atom_df['residue_name'] != 'HOH']
    #reset the indicies of the df
    atom_df = atom_df.reset_index(drop=True)

    #need to take out the ligand hetatms because they also have chain ID
    last_atom_idx = atom_df.loc[atom_df['atom_number']==last_atom_num].index
    atom_df = atom_df.iloc[:last_atom_idx[0]+1]

    #make a dictionary for translating chain id's into A,B,C,D... as Modeller likes
    id_dict = {}
    id_list = list(atom_df['chain_id'])
    alphabet_idx = 0
    for i in range(1,len(list(id_list))):
        if id_list[i] != id_list[i-1]:
            id_dict[id_list[i-1]] = chain_labels_l[alphabet_idx] 
            alphabet_idx = alphabet_idx + 1
        if (i == len(list(id_list))-1):
           id_dict[id_list[i]] = chain_labels_l[alphabet_idx]

    #now change atom_df to have labels modeller prefers
    atom_df['chain_id'] = atom_df['chain_id'].apply(lambda x: id_dict[x])

    return atom_df, others_df, id_dict

def get_pdb_seq(pdb_seq_data):
    pdb_seq = ""
    seq_data = pdb_seq_data[3:]
    for line in seq_data:
        line = line.rstrip()
        pdb_seq = pdb_seq + line
    #remove the break character '/'
    #pdb_seq = re.sub('/','',pdb_seq)
    return pdb_seq

def get_full_seq(fasta_data):
    full_seq = ''
    for line in fasta_data:
        if not(line[0] == '>'):
            line = line.rstrip() #I could have just messed up some crap with this
            full_seq = full_seq + line
        if (line[0] == '>'): #if have new chain
            full_seq = full_seq + '/' #to indicate chain break

    #get rid of beginning '/'        
    if full_seq[0] == '/':
        full_seq = full_seq[1:] #get rid of beginning '/'

    return full_seq + '*'

def trim_full_seq(pdb_seq, full_seq):
    pdb_letters = pdb_seq
    #trim off the front
    for i in range(len(full_seq)):
        if (full_seq[i:i+5] == pdb_letters[0:5]):
            full_seq = full_seq[i:]
            break #needed so proper length
    #trim off the back (make sure * is not present or will mess it up)
    for j in reversed(range(len(full_seq)+1)):
        if (full_seq[j-5:j] == pdb_letters[(len(pdb_letters)-5):]):
            full_seq = full_seq[0:j]
            break
    return full_seq

def make_one_letter(res_l):
    # Amino acid 3-to-1 letter code dictionary.
    aaDict = dict({'ALA': 'A', 
                    'ARG': 'R', 'ASN': 'N',
                    'ASP': 'D', 'CYS': 'C',
                    'GLN': 'Q', 'GLU': 'E', 
                    'GLY': 'G', 'HIS': 'H',
                    'ILE': 'I', 'LEU': 'L', 
                    'LYS': 'K', 'MET': 'M', 
                    'PHE': 'F', 'PRO': 'P', 
                    'SER': 'S', 'THR': 'T', 
                    'TRP': 'W', 'TYR': 'Y', 
                    'VAL': 'V'})
    #make the lift one letter
    one_letter_code_seq = []
    for res in res_l:
        #check if it is a natural amino acid
        if res in list(aaDict):
            one_letter_code_seq.append(aaDict[res])
        else:
            one_letter_code_seq.append('.') #for HETATMS (unnatural aa's)

    #now make the list one string
    seq_str = ''.join(one_letter_code_seq)
    return seq_str

def insert_dashes(string, index, num):
    dash_str = ''
    for i in range(num):
        dash_str = dash_str + '-'
    return string[:index] + dash_str + string[index:]

#Add dashes where the missing residues are in the pdb_seq
def add_gaps(seq_before,seq_after,full_seq,pdb_seq):
    #first turn the expressiongs into re, this works nicely because '.' is the wildcard character
    seq_before_re = re.compile(seq_before)
    seq_after_re = re.compile(seq_after)
    merged_seq_re = re.compile(seq_before + seq_after)

    full_seq_no_dash = full_seq.replace('-','') ##why did i make the full_seq no dash and not pdb_seq instead?????

    #find where gap begins, need to add back length because gives idx at beginning of str
    begin_gap_idx = re.search(seq_before_re,full_seq_no_dash).start() + len(seq_before)
    #begin_gap_idx = full_seq.find(seq_before) + len(seq_before) #probably wrong

    # find where gap ends
    end_gap_idx = re.search(seq_after_re,full_seq_no_dash).start()
    #end_gap_idx = full_seq.find(seq_after)

    #compute number of dashed
    num_dashes = end_gap_idx - begin_gap_idx
    #find where to insert gap in pdb sequence
    insert_idx = re.search(merged_seq_re,pdb_seq).start() + len(seq_before)
    #insert_idx = pdb_seq.find(seq_before + seq_after) + len(seq_before)

    #make new pdb_seq
    pdb_seq = insert_dashes(pdb_seq,insert_idx,num_dashes)

    return pdb_seq

def split_sequence(seq):
    return re.findall('.{1,75}', seq)

def break_into_chains(seq):
    return seq.split('/')

def trim_ligand_seq(hetatm_seq):
    ligand_str = ''
    seq = hetatm_seq[0:-1] #take off the final *
    new_seq = seq
    for j in reversed(range(len(seq))):
        #if not(seq[j] == '/' or seq[j] == '.'):
        if (seq[j] in chain_labels_l):
            new_seq = seq[0:j+1]
            ligand_str = seq[j+1:]
            break
    new_seq = new_seq + '*'
    return new_seq, ligand_str

# function to take in the others_df and return a list the missing res
def find_missing_res_l(others_df):
    # get df that has only the lines that start with 'REMARK'
    remark_df = others_df.loc[others_df['record_name']=='REMARK']
    # turn this data into a list of lists, which contain strings
    remark_l = list(remark_df['entry'].str.split())
    # find the data for all REMARK 465
    missing_res_ls = []
    new_missing_res_ls = []
    for l in remark_l:
        if (l[0] == '465' and len(l)>=2):
            missing_res_ls.append(l[1:])
    #exit if can't find any missing residues
    if (len(missing_res_ls) == 0):
        print("cannot find any missing residues")
        sys.exit(2)
    # get only the lines that contain the missing residues
    for i in range(len(missing_res_ls)):
        if (not(len(missing_res_ls[i]) == 0)):
            #take all lines after '465 M RES C SSSEQI'
            if (((missing_res_ls[i])[0]) == 'M'):
                #call it something different so it doesn't mess up the size
                new_missing_res_ls = missing_res_ls[i+1:]
    return new_missing_res_ls

# make a dictionary with chain id's as the keys and list of missing res as the values
def make_missing_res_dict(others_df, id_dict):

    # first, get the missing_res_ls
    missing_res_ls = find_missing_res_l(others_df)
    # initialize dict with empty lists as values
    missing_res_dict = {k:[] for k in chain_labels_l}
    # populate the dict based on chain id
    for l in missing_res_ls:
        missing_res_dict[id_dict[l[1]]].append([l[0],l[2]])
    # now seperate into loops, based on if res_num is sequential
    missing_res_dict_tmp = {k:[] for k in chain_labels_l}
    #loop over dictionary
    for key in missing_res_dict: 
        start_idx = 0
        for i in range(1,len(missing_res_dict[key])):
            res_num = int(missing_res_dict[key][i][1])
            prev_res_num = int(missing_res_dict[key][i-1][1])
            # if res_nums are not sequential, then end of loop
            if (res_num != prev_res_num + 1):
                missing_res_dict_tmp[key].append(missing_res_dict[key][start_idx:i])
                start_idx = i
            # if last residue, then end of loop
            if (i == len(missing_res_dict[key])-1):
                missing_res_dict_tmp[key].append(missing_res_dict[key][start_idx:i+1])
    missing_res_dict = missing_res_dict_tmp
    return missing_res_dict

# make a dictionary of only the residue numbers from missing_res_dict
def make_res_num_dict(missing_res_dict):

    # initialize the dict values to empty lists
    missing_res_num_dict = {k:[] for k in chain_labels_l}

    for key in missing_res_dict:
        for missing_loop_l in missing_res_dict[key]:
            loop_res_num_l = []
            for res in missing_loop_l:
                res_num = int(res[1])
                loop_res_num_l.append(res_num)

            missing_res_num_dict[key].append(loop_res_num_l)

    return missing_res_num_dict
            
# make a dictionary of only the residue letters (1 letter code) from missing_res_dict
def make_res_letter_dict(missing_res_dict):

    # initialize the dict values to empty lists
    missing_res_letter_dict = {k:[] for k in chain_labels_l}

    for key in missing_res_dict:
        for missing_loop_l in missing_res_dict[key]:
            loop_res_letter_l = []
            for res in missing_loop_l:
                res_letter = res[0]
                loop_res_letter_l.append(res_letter)

        missing_res_letter_dict[key].append(loop_res_letter_l)

    # turn 3 letter code into one letter code, and also turn list into str
    one_letter_dict = {k:[] for k in chain_labels_l}
    for key in missing_res_letter_dict:
        for l in missing_res_letter_dict[key]:
            seq_str = make_one_letter(l)
            one_letter_dict[key].append(seq_str)

    return one_letter_dict

#code to turn list of chain seqs into a dictionary where keys are chain ids 
def make_chain_dict(seq_l):
    chain_dict = {k:[] for k in chain_labels_l}

    letter_idx=0
    for chain in seq_l:
        chain_letter = chain_labels_l[letter_idx]
        chain_dict[chain_letter] = chain
        letter_idx = letter_idx + 1

    return chain_dict

#function to get a list of residues that flank the gaps
def get_flank_res_num_l(missing_res_loops):
    flank_res_l = []
    for loop in missing_res_loops:
        #find the first and last residue of the loop, then go outside by one
        flank_res_l.append([loop[0]-1,loop[-1]+1])
    return flank_res_l

def get_flank_seq(atom_df, flank_res_num, key):

    first_res_num = flank_res_num[0]
    second_res_num = flank_res_num[1]

    #need to only get rows with the correct chain id

    chain_df = atom_df.loc[atom_df['chain_id']==key]
    #reset the index to 0
    chain_df = chain_df.reset_index(drop=True)

    #l = flank_res_num # for convenience of writing

    #get the first atom_number of the pandas df that this res is at
    try:
        first_atom_idx = (chain_df.loc[chain_df['residue_number'] ==  first_res_num,'atom_number'].index)[0]
        second_atom_idx = (chain_df.loc[chain_df['residue_number'] == second_res_num,'atom_number'].index)[-1]
    #if at the end of a chain    
    except IndexError:
        return '','','',''

    #make sure they are not at the end of the chain:
    if (first_atom_idx == [] or second_atom_idx == []):
        return '','','',''

    # when selecting residues, need to know if at end of chain, don't want to call out of bounds idx
    if (first_atom_idx-40 < 0):
        lower_bound = 0
    else:
        lower_bound = first_atom_idx-40

    if (second_atom_idx+40 >= chain_df.shape[0]):
        upper_bound = chain_df.shape[0] - 1
    else:
        upper_bound = second_atom_idx+40

    # get lists of the residues (and their numbers) before and after the missing residues
    res_before_l = list(chain_df.loc[lower_bound:first_atom_idx]['residue_name'])
    resnum_before_l = list(chain_df.loc[lower_bound:first_atom_idx]['residue_number'])

    res_after_l = list(chain_df.loc[second_atom_idx:upper_bound]['residue_name'])
    resnum_after_l = list(chain_df.loc[second_atom_idx:upper_bound]['residue_number'])

    # add the residue names to the list, and make sure only once per residue
    res_before = []
    res_before.append(res_before_l[0])
    for i in range(1,len(res_before_l)):
        if ((not(resnum_before_l[i] == resnum_before_l[i-1]))): #checking if same aa, is diff res
            res_before.append(res_before_l[i])

    # add the residue names to the list, and make sure only once per residue
    res_after = []
    res_after.append(res_after_l[0])
    for i in range(1,len(res_after_l)):
        if ((not(resnum_after_l[i] == resnum_after_l[i-1]))):
            res_after.append(res_after_l[i])

    merged_seq = res_before + res_after
    gap_res = [res_before[-1],res_after[0]]

    return merged_seq, gap_res, res_before, res_after

if __name__ == "__main__":
    
    main()


#!/usr/bin/env python
# File name: chop_chain_joiner.py
# Author: Matt Robinson
# Date created: 5/24/2017
# Date last modified: 5/24/2017
# Python Version: 3.6
"""
Description:
This script 'fills in' missing residues in PDB files and creates a fixed PDB file. 
In order to fill in the gaps, Modeller is used to create a homology model with the 
original PDB file serving as a template and the full sequence serving as the target 
sequence to model. 

The process of doing this is as follows:

1. Call make_seq.py to extract the sequence from the original PDB file with missing residues.

2. Call make_alignment.py to create an alignment file, alignment.ali, between the original PDB structure
with missing residues and the full fasta sequence (usually available on PDB website). 

3. Call make_model.py to create the actual homology model.

Please see the headers of each of these scripts for more specific information.

Usage: python chain_joiner.py -p pdbfile.pdb -f fastafile.fasta [options]

For example, to make a loop model of PDB code 1qg8, I would call it as: 
'python chain_joiner.py -p 1qg8.pdb -f 1qg8_full_seq.fasta -a'

Input Arguments:
[optional]
-a, --automodel
    The simplest method for simple comparitive modeling. Will not give 
    great results but suggested when many chain breaks are present. [default: True]

-f, --fixed_automodel
    Builds an automodel and keeps the non-missing residues fixed, 
    whereas they can move in the other methods. [default: False]

-l, --loopmodel 
    Builds a model by refining the loop with the missing residues.
    Suggested when have one small chain break in the PDB. [default: False]

Output: A set of PDB files (number depends on the chosen method)
"""

import argparse
import sys
import os

# import modules
# from chain_joiner import make_seq
# from chain_joiner import make_alignment
# from chain_joiner import make_model

from chain_joiner import make_seq
from chain_joiner import make_alignment
from chain_joiner import make_model

def main():

    parser = argparse.ArgumentParser(
        prog='make_model.py',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=
    """
    This script 'fills in' missing residues in PDB files and creates a fixed PDB file. 
    In order to fill in the gaps, Modeller is used to create a homology model with the 
    original PDB file serving as a template and the full sequence serving as the target 
    sequence to model. 

    The process of doing this is as follows:

    1. Call make_seq.py to extract the sequence from
       the original PDB file with missing residues.

    2. Call make_alignment.py to create an alignment file, alignment.ali, 
       between the original PDB structure with missing residues and the full
       fasta sequence (usually available on PDB website). 

    3. Call make_model.py to create the actual homology model.

    Please see the headers of each of these scripts for more specific information.

    Usage: python chain_joiner.py -p pdbfile.pdb -f fastafile.fasta [options]

    For example, to make a loop model of PDB code 1qg8, I would call it as: 
    'python chain_joiner.py -p 1qg8.pdb -f 1qg8_full_seq.fasta -a'

    @author: Matt Robinson, matthew.robinson@yale.edu
             William L. Jorgensen lab, Yale University
    
    REQUIREMENTS:
    Preferably Anaconda python 3 with following modules:
        argparse
        modeller    
    """
    )

    parser.add_argument(
        "-p", "--pdb", help="path of the pdb file with .pdb file descriptor")
    parser.add_argument(
        "-f", "--fasta", help="path of the fasta file with .fasta file descriptor")
    parser.add_argument(
        "-a", "--automodel", help="the simplest method for simple comparitive modeling", action="store_true")
    parser.add_argument(
        "-fm", "--fixed_automodel", help="builds an automodel and keeps the non-missing residues fixed", action="store_true")
    parser.add_argument(
        "-l", "--loopmodel", help="builds a model by refining the loop with the missing residues", action="store_true")

    args = parser.parse_args()

    # join the chains
    join_chains(args.pdb, args.fasta, args.automodel, args.fixed_automodel, args.loopmodel)

def join_chains(pdb_file, fasta_file, a, fm, l):

    # Get the PDB id from the file
    pdb_id = os.path.splitext(os.path.basename(pdb_file))[0]

    # get the sequence from the PDB file:
    make_seq.get_sequence(pdb_file)

    # make the alignment file:
    make_alignment.align(pdb_file, pdb_id + ".seq", fasta_file)

    # make the model
    make_model.model(pdb_file, a, fm, l)

    # make a folder for the output
    dir_name = './' + pdb_id +'_output/'
    os.makedirs(dir_name)

    # get a list of all output files in the working directory
    output_files = [filename for filename in os.listdir('.') if filename.startswith(pdb_id)]
    # remove the folder name
    if (pdb_id + '_output') in output_files:
        output_files.remove(pdb_id + '_output')

    # mv these files to the output folder
    for file in output_files:
        try:
            os.system('mv ' + file + ' ./' + pdb_id + '_output/')   
        except:
            pass

if __name__ == "__main__":

    main()


#!/usr/bin/env python
# File name: make_seq.py
# Author: Matt Robinson
# Date created: 5/23/2017
# Date last modified: 9/05/2017
# Python Version: 3.6
"""
Description:
This script extracts the sequence from a PDB file with missing residues. 
The output is a sequence file of the form [pdb_filename].seq

Usage: python make_pdb_seq.py -p pdbfile.pdb
"""
import sys
import os
import argparse
from modeller import *

def main():

    parser = argparse.ArgumentParser(
        prog='make_seq.py',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=
    """
    This script extracts the sequence from a PDB file with missing residues.
    Output is a sequence file of the form [pdb_filename].seq

    @author: Matt Robinson, matthew.robinson@yale.edu
             William L. Jorgensen lab, Yale University

    Usage: python make_seq.py -p pdbfile.pdb
    
    Example Usage: python make_seq.py -p 1qg8.pdb
    
    REQUIREMENTS:
    Preferably Anaconda python 3 with following modules:
        argparse
        modeller    
    """
    )

    parser.add_argument(
        "-p", "--pdb", help="full path of the pdb file with .pdb file descriptor")

    args = parser.parse_args()

    # call the main function
    get_sequence(args.pdb)

def get_sequence(pdb_file):

    #Get the PDB id from the file (just strips off extension)
    pdb_id = os.path.splitext(os.path.basename(pdb_file))[0]

    # Get the sequence of the PDB file, and write to an alignment file
    code = pdb_file

    e = environ()

    # directories for input atom files
    e.io.atom_files_directory = ['.', '../atom_files']
    # Read in HETATM records from template PDBs
    e.io.hetatm = True

    m = model(e, file=code)
    aln = alignment(e)
    aln.append_model(m, align_codes=code)
    aln.write(file=pdb_id+'.seq')

if __name__ == "__main__":

    main()




from setuptools import setup, find_packages

setup(
  name = 'chain_joiner',
  packages = ['chain_joiner'], # this must be the same as the name above
  version = '0.11',
  description = 'A package to automatically fix chain breaks in PDB files using MODELLER',
  author = 'Matthew C. Robinson',
  author_email = 'matthew.robinson@yale.edu',
  url = 'https://github.com/mc-robinson/chain_joiner', # use the URL to the github repo,
  keywords = [], # arbitrary keywords
  classifiers = [
	'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering',
	],
  install_requires=['biopandas'],
  entry_points={
        'console_scripts': [
            'join_chains=chain_joiner.chain_joiner:main',
	    'make_seq=chain_joiner.make_seq:main',
	    'make_alignment=chain_joiner.make_alignment:main',
            'make_model=chain_joiner.make_model:main',
            'chain_joiner_online=chain_joiner.run_chain_joiner_online:main'
        ],
    },

)



#!/usr/bin/env python
# File name: run_chain_joiner.py
# Author: Matt Robinson
# Date created: 06/09/2017
# Date last modified: 06/09/2017
# Python Version: 3.6
"""
This program uses wget to obtain the files needed for chain_joiner from the PDB website.
It then runs chain_joiner in automodel mode. 

usage: python run_chain_joiner.py pdb_code
"""

import os
import sys
import subprocess
import argparse

from chain_joiner import chain_joiner

def main():
    #create parser object
    parser = argparse.ArgumentParser(
        prog='run_chain_joiner_online.py',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=
    """
    SCRIPT TO FIX MISSING RESIDUES IN PDB FILES.
    This version obtains the PDB and FASTA files 
    for the user directly from the PDB website.

    @author: Matthew Robinson, matthew.robinson@yale.edu
    @author: William L. Jorgensen Lab, Yale University 

    Usage: python run_chain_joiner_online.py -p 1qg8 -a
    
    REQUIREMENTS:
    Preferably Anaconda python 3 environment with following modules:    
        Modeller
        pandas 
        biopandas
        argparse
    """
    )

    # defining arguments for parser object
    parser.add_argument(
        "-p", "--pdb_id", help="pdb four letter code for the protein of interest")
    parser.add_argument(
        "-a", "--automodel", help="the simplest method for simple comparitive modeling", action="store_true")
    parser.add_argument(
        "-fm", "--fixed_automodel", help="builds an automodel and keeps the non-missing residues fixed", action="store_true")
    parser.add_argument(
        "-l", "--loopmodel", help="builds a model by refining the loop with the missing residues", action="store_true")

    #parse the arguments from standard input
    args = parser.parse_args()

    # call the model function
    model(args.pdb_id, args.automodel, args.fixed_automodel, args.loopmodel)

def model(pdb_id, a, fm, l):

    # make url's for wget
    pdb_url = 'https://files.rcsb.org/download/' + pdb_id + '.pdb'
    # fasta_url = 'http://www.rcsb.org/pdb/download/viewFastaFiles.do?structureIdList=' + pdb_id + '&compressionType=uncompressed'
    fasta_url = 'http://www.rcsb.org/pdb/download/viewFastaFiles.do?structureIdList=' + pdb_id

    #make file names
    pdb_fn = pdb_id + '.pdb'
    fasta_fn = pdb_id + '.fasta'

    #get the files from the pdb_database
    os.system('wget --no-check-certificate -O ' + pdb_fn + ' ' + pdb_url)
    os.system('wget -O ' + fasta_fn +'.gz ' + fasta_url + ' && gzip -d '+ fasta_fn +'.gz')
    #os.system('wget -q -O - ' + fasta_url + ' | gzip -d > ' + fasta_fn)
    #os.system('gzip -d ' + fasta_fn + '.gz')

    # run chain joiner
    chain_joiner.join_chains(pdb_fn, fasta_fn, a, fm, l )

if __name__ == "__main__":

    main()